@@ -10,19 +10,16 @@ Classes
 =======
 
 """
-
 from collections import deque
 from threading import RLock
-import os
-import time
 
 import pyuv
 import six
 
 from .events import EventEmitter
 from .process import Process
-from .sync import increment, add, sub, atomic_read, compare_and_swap
-
+from .state import ProcessState, ProcessTracker
+from .sync import increment
 
 class Manager(object):
     """ Manager - maintain process alive
@@ -37,17 +34,17 @@ class Manager(object):
     The design is pretty simple. The manager is running on the default
     event loop and listening on events. Events are sent when a process
     exit or from any method call. The control of a manager can be
-    extended by adding controllers on startup. For example gaffer
-    provides a controller allowing you to control processes via HTTP.
+    extended by adding apps on startup. For example gaffer
+    provides an application allowing you to control processes via HTTP.
 
-    Running a controller is done like this::
+    Running an application is done like this::
 
-        # initialize the controller with the default loop
+        # initialize the application with the default loop
         loop = pyuv.Loop.default_loop()
         m = Manager(loop=loop)
 
-        # start the controller
-        m.start(controllers=[HttpHandler])
+        # start the application
+        m.start(apps=[HttpHandler])
 
         .... # do smth
 
@@ -73,14 +70,17 @@ class Manager(object):
         # by default we run on the default loop
         self.loop = loop or pyuv.Loop.default_loop()
 
-        # setup wakeup event
+        # wakeup ev for internal signaling
         self._wakeup_ev = pyuv.Async(self.loop, self._on_wakeup)
 
         # initialize the emitter
         self._emitter = EventEmitter(self.loop)
 
+        # initialize the process tracker
+        self._tracker = ProcessTracker(self.loop)
+
         # initialize some values
-        self.controllers = None
+        self.apps = None
         self.started = False
         self._stop_ev = None
         self.max_process_id = 0
@@ -88,18 +88,22 @@ class Manager(object):
         self.running = {}
         self.channel = deque()
         self._updates = deque()
-        self._signals = deque()
+        self._signals = []
 
+        self.stopping= False
         self.stop_cb = None
         self.restart_cb = None
         self._lock = RLock()
 
-    def start(self, controllers=[]):
+    def start(self, apps=[]):
         """ start the manager. """
-        self.controllers = controllers
+        self.apps = apps
+
+        # start the process tracker
+        self._tracker.start()
 
         # start contollers
-        for ctl in self.controllers:
+        for ctl in self.apps:
             ctl.start(self.loop, self)
 
         self.started = True
@@ -122,6 +126,7 @@ class Manager(object):
         self._signals.append("STOP")
         self.wakeup()
 
+
     def restart(self, callback=None):
         """ restart all processes in the manager. This function is
         threadsafe """
@@ -206,6 +211,9 @@ class Manager(object):
           redirect stdoutt & stderr and stdout events will be labeled "a"
         - **redirect_input**: Boolean (False is the default). Set it if
           you want to be able to write to stdin.
+        - **graceful_timeout**: graceful time before we send a  SIGKILL
+          to the process (which definitely kill it). By default 30s.
+          This is a time we let to a process to exit cleanly.
         """
 
         with self._lock:
@@ -222,9 +230,7 @@ class Manager(object):
 
             self._publish("create", name=name)
             if start:
-                self._publish("start", name=name)
-                self._publish("proc.%s.start" % name, name=name)
-                self._spawn_processes(state)
+                self.start_process(name)
 
     def update_process(self, name, cmd, **kwargs):
         """ update a process information.
@@ -244,7 +250,18 @@ class Manager(object):
                 del kwargs['start']
 
             self._publish("update", name=name)
-            self._spawn_processes(state)
+            self._manage_processes(state)
+
+
+    def start_process(self, name):
+        with self._lock:
+            if name in self.processes:
+                state = self.processes[name]
+                self._publish("start", name=name)
+                self._publish("proc.%s.start" % name, name=name)
+                self._manage_processes(state)
+            else:
+                raise KeyError("%s not found")
 
     def stop_process(self, name_or_id):
         """ stop a process by name or id
@@ -254,14 +271,23 @@ class Manager(object):
         process id is givien, only the process with this id will be
         stopped """
 
-        if isinstance(name_or_id, six.string_types):
-            stop_func = self._stop_byname_unlocked
-        else:
-            stop_func = self._stop_byid_unlocked
-
-        # really stop the process
         with self._lock:
-            stop_func(name_or_id)
+            # stop all processes of the template name
+            if isinstance(name_or_id, six.string_types):
+                self._stop_processes(name_or_id)
+            else:
+                # stop a process by its internal pid
+                self._stop_process(name_or_id)
+
+
+    def restart_process(self, name):
+        """ restart a process """
+        with self._lock:
+            if name not in self.processes:
+                raise KeyError("%r not found" % name)
+
+            state = self.get_process_state(name)
+            self._restart_processes(state)
 
     def remove_process(self, name):
         """ remove the process and its config from the manager """
@@ -270,10 +296,21 @@ class Manager(object):
             if name not in self.processes:
                 raise KeyError("%r not found" % name)
 
-            self._stop_byname_unlocked(name)
+            # stop all processes
+            self._stop_processes(name)
+            # remove it the list
             del self.processes[name]
+            # notify other that this template has been deleted
             self._publish("delete", name=name)
 
+    def manage_process(self, name):
+        with self._lock:
+            if name not in self.processes:
+                raise KeyError("%r not found" % name)
+            state = self.processes[name]
+            self._manage_processes(state)
+
+
     def get_process_info(self, name):
         """ get process info """
         with self._lock:
@@ -286,6 +323,21 @@ class Manager(object):
             return info
 
     def get_process_status(self, name):
+        """ return the process status::
+
+            {
+              "active":  str,
+              "running": int,
+              "max_processes": int
+            }
+
+        - **active** can be *active* or *stopped*
+        - **running**: the number of actually running OS processes using
+          this template.
+        - **max_processes**: The maximum number of processes that should
+          run. It is is normally the same than the **runnin** value.
+
+        """
         with self._lock:
             if name not in self.processes:
                 raise KeyError("%r not found" % name)
@@ -316,54 +368,27 @@ class Manager(object):
         """ get stats changes on a process template or id
         """
         with self._lock:
-            if isinstance(name_or_id, int):
-                try:
+            try:
+                if isinstance(name_or_id, int):
                     return self.running[name_or_id].monitor(listener)
-                except KeyError:
-                    raise KeyError("%s not found" % name_or_id)
-            else:
-                if name_or_id not in self.processes:
-                    raise KeyError("%r not found" % name_or_id)
-
-                state = self.processes[name_or_id]
-                return state.monitor(listener)
+                else:
+                    state = self.processes[name_or_id]
+                    return state.monitor(listener)
+            except KeyError:
+                raise KeyError("%s not found" % name_or_id)
 
     def unmonitor(self, name_or_id, listener):
         """ get stats changes on a process template or id
         """
         with self._lock:
-            if isinstance(name_or_id, int):
-                try:
+            try:
+                if isinstance(name_or_id, int):
                     return self.running[name_or_id].unmonitor(listener)
-                except KeyError:
-                    raise KeyError("%s not found" % name_or_id)
-            else:
-                if name_or_id not in self.processes:
-                    raise KeyError("%r not found" % name_or_id)
-
-                state = self.processes[name_or_id]
-                return state.unmonitor(listener)
-
-    def manage_process(self, name):
-        with self._lock:
-            state = self.get_process_state(name)
-            state.stopped = False
-            self._manage_processes(state)
-
-    def start_process(self, name):
-        self._publish("start", name=name)
-        self._publish("proc.%s.start" % name, name=name)
-        self.manage_process(name)
-
-    def reap_process(self, name):
-        with self._lock:
-            self._reap_processes(self.get_process_state(name))
-
-    def restart_process(self, name):
-        """ restart a process """
-        with self._lock:
-            state = self.get_process_state(name)
-        self._restart_processes(state)
+                else:
+                    state = self.get_process_state(name_or_id)
+                    return state.unmonitor(listener)
+            except KeyError:
+                raise KeyError("%s not found" % name_or_id)
 
     def ttin(self, name, i=1):
         """ increase the number of system processes for a state. Change
@@ -403,19 +428,11 @@ class Manager(object):
             except KeyError:
                 pass
 
-
     # ------------- general purpose utilities
 
     def wakeup(self):
         self._wakeup_ev.send()
 
-    def update_state(self, name):
-        """ update the state. When the event loop is idle, the state is
-        read and processes in the state managed """
-
-        self._updates.append(name)
-        self.wakeup()
-
     def get_process_state(self, name):
         if name not in self.processes:
             return
@@ -429,55 +446,120 @@ class Manager(object):
 
     # ------------- private functions
 
-    def _shutdown(self, handle):
-
-        def clean_cb(h):
-            if h.active:
-                h.close()
-
-        self.loop.walk(clean_cb)
-        handle.stop()
-
-
-    def _stop(self):
-        # stop all processes
-        self.stop_processes()
-
-        if self.processes:
-            # graceful shutdown let a chance to unstopped process to
-            # close cleanly
-            self._shutdown_h = pyuv.Timer(self.loop)
-            self._shutdown_h.start(self._shutdown, 0.1, 0.0)
-
+    def _shutdown(self):
         with self._lock:
-            # stop controllers
-            for ctl in self.controllers:
+            # stop the applications.
+            for ctl in self.apps:
                 ctl.stop()
 
-
             # we are now stopped
             self.started = False
 
-            if self.stop_cb:
+            # close all handles
+            #def walk_cb(h):
+            #    if h.active:
+            #        h.close()
+            #self.loop.walk(walk_cb)
+
+            # if there any stop callback, excute it
+            if self.stop_cb is not None:
                 self.stop_cb(self)
                 self.stop_cb = None
 
+    def _stop(self):
+        # stop should be synchronous. We need to first stop the
+        # processes and let the applications know about it. It is
+        # actually done by setting on startup a timer waiting that all
+        # processes have stopped to run. Then applications are stopped.
+
+        self.stopping = True
+
+        # stop all processes
+        with self._lock:
+            for name in self.processes:
+                self._stop_processes(name)
+
+            self._tracker.on_done(self._shutdown)
+
     def _restart(self):
         with self._lock:
-            to_restart = [state for name, state in self.processes.items()]
+            # on restart we first restart the applications
+            for app in self.apps:
+                app.restart()
 
-        for state in to_restart:
-            if state is not None:
+            # then we restart the processes
+            for name, state in self.processes.items():
                 self._restart_processes(state)
 
-        with self._lock:
-            for ctl in self.controllers:
-                ctl.restart()
-
-            if self.restart_cb:
+            # if any callback has been set, run it
+            if self.restart_cb is not None:
                 self.restart_cb(self)
                 self.restart_cb = None
 
+    def _stop_processes(self, name):
+        """ stop all processes in a template """
+        if name not in self.processes:
+            print("return")
+            return
+
+        # get the template
+        state = self.processes[name]
+        if state.stopped:
+            return
+        state.stopped = True
+
+        # notify others that all processes of the templates are beeing
+        # stopped.
+        self._publish("stop", name=name)
+        self._publish("proc.%s.stop" % name, name=name)
+
+        # stop the flapping detection.
+        if state.flapping_timer is not None:
+            state.flapping_timer.stop()
+
+        # iterrate over queued processes.
+        while True:
+            try:
+                p = state.dequeue()
+            except IndexError:
+                break
+
+            # notify  other that the process is beeing stopped
+            self._publish("proc.%s.stop_pid" % p.name, name=p.name,
+                    pid=p.id)
+
+            # remove the pid from the running processes
+            if p.id in self.running:
+                self.running.pop(p.id)
+
+            # stop the process
+            p.stop()
+
+            # track this process to make sure it's killed after the
+            # graceful time
+            self._tracker.check(p, state.graceful_timeout)
+
+    def _stop_process(self, pid):
+        """ stop a process bby id """
+
+        if pid not in self.running:
+            return
+
+        # remove the process from the running processes
+        p = self.running.pop(pid)
+        state = self.processes[p.name]
+        state.remove(p)
+
+        # stop the process
+        p.stop()
+
+        # track this process to make sure it's killed after the
+        # graceful time
+        self._tracker.check(p, state.graceful_timeout)
+
+        # notify  other that the process is beeing stopped
+        self._publish("proc.%s.stop_pid" % p.name, name=p.name, pid=pid)
+
     def _spawn_process(self, state):
         """ spawn a new process and add it to the state """
         # get internal process id
@@ -496,49 +578,6 @@ class Manager(object):
         self._publish("proc.%s.spawn" % p.name, name=p.name, pid=pid,
                 detached=p.detach)
 
-    def _stop_byname_unlocked(self, name):
-        """ stop a process by name """
-        if name not in self.processes:
-            return
-
-        self._publish("stop", name=name)
-        self._publish("proc.%s.stop" % name, name=name)
-
-        state = self.processes[name]
-        state.stopped = True
-
-        if state.flapping_timer is not None:
-            state.flapping_timer.stop()
-
-        while True:
-            try:
-                p = state.dequeue()
-            except IndexError:
-                break
-
-            # race condition, we need to remove the process from the
-            # running pid now.
-            if p.id in self.running:
-                self.running.pop(p.id)
-            p.stop()
-
-
-    def _stop_byid_unlocked(self, pid):
-        """ stop a process bby id """
-        if pid not in self.running:
-            return
-
-        p = self.running.pop(pid)
-
-        self._publish("proc.%s.stop_pid" % p.name, name=p.name, pid=pid)
-
-        # remove the process from the running processes
-        state = self.processes[p.name]
-        state.remove(p)
-
-        # finally stop the process
-        p.stop()
-
     def _spawn_processes(self, state):
         """ spawn all processes for a state """
         num_to_start = state.numprocesses - len(state.running)
@@ -547,45 +586,41 @@ class Manager(object):
 
     def _reap_processes(self, state):
         diff = len(state.running) - state.numprocesses
-
         if diff > 0:
             for i in range(diff):
-                p = state.dequeue()
+                # remove the process from the running processes
+                try:
+                    p = state.dequeue()
+                except IndexError:
+                    return
+
+                # remove the pid from the running processes
+                if p.id in self.running:
+                    self.running.pop(p.id)
+
+                # stop the process
+                p.stop()
+
+                # track this process to make sure it's killed after the
+                # graceful time
+                self._tracker.check(p, state.graceful_timeout)
+
+                # notify others that the process is beeing reaped
                 self._publish("proc.%s.reap" % p.name, name=p.name,
                     pid=p.id)
-                p.stop()
 
     def _manage_processes(self, state):
         if len(state.running) < state.numprocesses:
             self._spawn_processes(state)
         self._reap_processes(state)
 
-
     def _restart_processes(self, state):
-        # disable automatic process management while we are stopping
-        # the processes
-        state.numprocesses = 0
+        # first launch new processes
+        for i in range(state.numprocesses):
+            self._spawn_process(state)
 
-        # stop the processes, we need to lock here
-        with self._lock:
-            while True:
-                try:
-                    p = state.dequeue()
-                except IndexError:
-                    break
-
-                # race condition, we need to remove the process from the
-                # running pid now.
-                if p.id in self.running:
-                    self.running.pop(p.id)
-
-                p.stop()
-
-        # reset the number of processes
-        state.reset()
-
-        # start the processes the next time
-        self.manage_process(state.name)
+        # then reap useless one.
+        self._manage_processes(state)
 
     def _check_flapping(self, state):
         if not state.flapping:
@@ -625,217 +660,43 @@ class Manager(object):
     # ------------- events handler
 
     def _on_wakeup(self, handle):
-        sig = None
-        try:
-            sig = self._signals.popleft()
-        except IndexError:
-            pass
-        if sig is not None:
-            sig = sig.upper()
-            if sig == "STOP":
-                handle.close()
-                return self._stop()
-            elif sig == "RESTART":
-                return self._restart()
+        sig = self._signals.pop(0) if len(self._signals) else None
+        if not sig:
+            return
 
-        if not len(self._updates):
-            pass
-
-        to_update = self._updates.popleft()
-        self.manage_process(to_update)
+        if sig == "STOP":
+            handle.close()
+            self._stop()
+        elif sig == "RESTART":
+            self._restart()
 
     def _on_exit(self, process, exit_status, term_signal):
-        """ exit callback returned when a process exit """
+        # exit callback called when a process exit
 
+        # notify other that the process exited
         self._publish("proc.%s.exit" % process.name, name=process.name,
                 pid=process.id, exit_status=exit_status,
                 term_signal=term_signal)
 
         with self._lock:
-            # remove the process from the running processes
+            # maybe uncjeck this process from the tracker
+            self._tracker.uncheck(process)
+
+            state = self.get_process_state(process.name)
+
+            # unexpected exit, remove the process from the list of
+            # running processes.
+
             if process.id in self.running:
                 self.running.pop(process.id)
+            state.remove(process)
 
-            # remove the running process from the process state
-            state = self.get_process_state(process.name)
+            # this template has been removed, return
             if not state:
                 return
-            state.remove(process)
-            # if not stopped we may need to restart a new process
-            if not state.stopped and self._check_flapping(state):
-                self._manage_processes(state)
 
-class FlappingInfo(object):
-    """ object to keep flapping infos """
-
-    def __init__(self, attempts=2, window=1., retry_in=7., max_retry=5):
-        self.attempts = attempts
-        self.window = window
-        self.retry_in = retry_in
-        self.max_retry = max_retry
-
-        # exit history
-        self.history = deque(maxlen = max_retry)
-        self.retries = 0
-
-    def reset(self):
-        self.history.clear()
-        self.retries = 0
-
-class ProcessState(object):
-    """ object used by the manager to maintain the process state """
-
-    DEFAULT_PARAMS = {
-            "group": None,
-            "args": None,
-            "uid": None,
-            "gid": None,
-            "cwd": None,
-            "detach": False,
-            "shell": False,
-            "redirect_output": [],
-            "redirect_input": False}
-
-    def __init__(self, name, cmd, **settings):
-        self.running = deque()
-        self.stopped = False
-        self.setup(name, cmd, **settings)
-
-    def setup(self, name, cmd, **settings):
-        self.name = name
-        self.cmd = cmd
-        self.settings = settings
-        self._numprocesses = self.settings.get('numprocesses', 1)
-
-        # set flapping
-        self.flapping = self.settings.get('flapping')
-        if isinstance(self.flapping, dict):
-            try:
-                self.flapping = FlappingInfo(**self.flapping)
-            except TypeError: # unknown value
-                self.flapping = None
-
-        self.flapping_timer = None
-
-        self.stopped = False
-
-    @property
-    def active(self):
-        return len(self.running) > 0
-
-    def __str__(self):
-        return "state: %s" % self.name
-
-    def make_process(self, loop, id, on_exit):
-        params = {}
-        for name, default in self.DEFAULT_PARAMS.items():
-            params[name] = self.settings.get(name, default)
-
-        os_env = self.settings.get('os_env', False)
-        if os_env:
-            env = params.get('env', {})
-            env.update(os.environ())
-            params['env'] = env
-
-        params['on_exit_cb'] = on_exit
-
-        return Process(loop, id, self.name, self.cmd, **params)
-
-    @property
-    def group(self):
-        return self.settings.get('group')
-
-    def __get_numprocesses(self):
-        return atomic_read(self._numprocesses)
-    def __set_numprocesses(self, n):
-        self._numprocesses = compare_and_swap(self._numprocesses, n)
-    numprocesses = property(__get_numprocesses, __set_numprocesses,
-            doc="""return the max numbers of processes that we keep
-            alive for this command""")
-
-    @property
-    def hup(self):
-        return self.settings.get('hup', False)
-
-    def reset(self):
-        self.numprocesses = self.settings.get('numprocesses', 1)
-        # reset flapping
-        if self.flapping and self.flapping is not None:
-            self.flapping.reset()
-
-    def ttin(self, i=1):
-        self._numprocesses = add(self._numprocesses, i)
-        return self._numprocesses
-
-    def ttou(self, i=1):
-        self._numprocesses = sub(self._numprocesses, i)
-        return self._numprocesses
-
-    def queue(self, process):
-        self.running.append(process)
-
-    def dequeue(self):
-        return self.running.popleft()
-
-    def remove(self, process):
-        try:
-            self.running.remove(process)
-        except ValueError:
-            pass
-
-    def list_processes(self):
-        return list(self.running)
-
-    def stats(self):
-        infos = []
-        lmem = []
-        lcpu = []
-        for p in self.running:
-            info = p.info
-            infos.append(info)
-            lmem.append(info['mem'])
-            lcpu.append(info['cpu'])
-
-        if 'N/A' in lmem:
-            mem, max_mem, min_mem = "N/A"
-        else:
-            max_mem = max(lmem)
-            min_mem = min(lmem)
-            mem = sum(lmem)
-
-        if 'N/A' in lcpu:
-            cpu, max_cpu, min_cpu = "N/A"
-        else:
-            max_cpu = max(lcpu)
-            min_cpu = min(lcpu)
-            cpu = sum(lcpu)
-
-        ret = dict(name=self.name, stats=infos, mem=mem, max_mem=max_mem,
-                min_mem=min_mem, cpu=cpu, max_cpu=max_cpu,
-                min_cpu=min_cpu)
-        return ret
-
-    def monitor(self, listener):
-        for p in self.running:
-            p.monitor(listener)
-
-    def unmonitor(self, listener):
-        for p in self.running:
-            p.unmonitor(listener)
-
-    def check_flapping(self):
-        f = self.flapping
-        if len(f.history) < f.attempts:
-            f.history.append(time.time())
-        else:
-            diff = f.history[-1] - f.history[0]
-            if diff > f.window:
-                f.reset()
-                f.history.append(time.time())
-            elif f.retries < f.max_retry:
-                increment(f.retries)
-                return False, True
-            else:
-                f.reset()
-                return False, False
-        return True, None
+            # eventually restart the process
+            if not state.stopped:
+                # manage the template, eventually restart a new one.
+                if self._check_flapping(state):
+                    self._manage_processes(state)
